from .boxplot import boxplot
from .distribution import distribution
from .pca import pca
from .permutation_test import permutation_test
from .roc_plot import roc_plot
from .scatter import scatter
from .scatterCI import scatterCI

__all__ = [
    "boxplot",
    "distribution",
    "pca",
    "permutation_test",
    "roc_plot",
    "scatter",
    "scatterCI"
]



import numpy as np


def knnimpute(X, k, verbose=False, print_interval=100):
    """kNN missing value imputation."""
    
    missing_mask = np.isnan(X)
    n_rows, n_cols = X.shape
        
    X_row_major = X.copy("C")
    if missing_mask.sum() != np.isnan(X_row_major).sum():
        X_row_major[missing_mask] = np.nan
    
    n_rows, n_cols = X.shape

    # matrix of mean squared difference between between samples
    D = np.ones((n_rows, n_rows), dtype="float32", order="C") * np.inf

    # we can cheaply determine the number of columns that two rows share
    # by taking the dot product between their finite masks
    observed_elements = np.isfinite(X_row_major).astype(int)
    n_shared_features_for_pairs_of_rows = np.dot(
        observed_elements,
        observed_elements.T)
    no_overlapping_features_rows = n_shared_features_for_pairs_of_rows == 0
    number_incomparable_rows = no_overlapping_features_rows.sum(axis=1)
    row_overlaps_every_other_row = (number_incomparable_rows == 0)
    row_overlaps_no_other_rows = number_incomparable_rows == n_rows
    valid_rows_mask = ~row_overlaps_no_other_rows
    valid_row_indices = np.where(valid_rows_mask)[0]

    # preallocate all the arrays that we would otherwise create in the
    # following loop and pass them as "out" parameters to NumPy ufuncs
    diffs = np.zeros_like(X_row_major)
    missing_differences = np.zeros_like(diffs, dtype=bool)
    valid_rows = np.zeros(n_rows, dtype=bool)
    ssd = np.zeros(n_rows, dtype=X_row_major.dtype)

    for i in valid_row_indices:
        x = X_row_major[i, :]
        np.subtract(X_row_major, x.reshape((1, n_cols)), out=diffs)
        np.isnan(diffs, out=missing_differences)

        # zero out all NaN's
        diffs[missing_differences] = 0

        # square each difference
        diffs **= 2

        observed_counts_per_row = n_shared_features_for_pairs_of_rows[i]

        if row_overlaps_every_other_row[i]:
            # add up all the non-missing squared differences
            diffs.sum(axis=1, out=D[i, :])
            D[i, :] /= observed_counts_per_row
        else:
            np.logical_not(no_overlapping_features_rows[i], out=valid_rows)

            # add up all the non-missing squared differences
            diffs.sum(axis=1, out=ssd)
            ssd[valid_rows] /= observed_counts_per_row[valid_rows]
            D[i, valid_rows] = ssd[valid_rows]
    
    min_dist=1e-6
    max_dist_multiplier=1e6
    
    D_finite_flat = D[np.isfinite(D)]
    if len(D_finite_flat) > 0:
        max_dist = max_dist_multiplier * max(1, D_finite_flat.max())
    else:
        max_dist = max_dist_multiplier
    # set diagonal of distance matrix to a large value since we don't want
    # points considering themselves as neighbors
    np.fill_diagonal(D, max_dist)
    D[D < min_dist] = min_dist  # prevents 0s
    D[D > max_dist] = max_dist  # prevents infinities
    
    X_result = X_row_major
    D = D
    effective_infinity = max_dist
    
    for i in range(n_rows):
        for j in np.where(missing_mask[i, :])[0]:
            distances = D[i, :].copy()

            # any rows that don't have the value we're currently trying
            # to impute are set to infinite distances
            distances[missing_mask[:, j]] = effective_infinity
            neighbor_indices = np.argsort(distances)
            neighbor_distances = distances[neighbor_indices]

            # get rid of any infinite distance neighbors in the top k
            valid_distances = neighbor_distances < effective_infinity
            neighbor_distances = neighbor_distances[valid_distances][:k]
            neighbor_indices = neighbor_indices[valid_distances][:k]

            weights = 1.0 / neighbor_distances
            weight_sum = weights.sum()

            if weight_sum > 0:
                column = X[:, j]
                values = column[neighbor_indices]
                X_result[i, j] = np.dot(values, weights) / weight_sum
    return X_result



import numpy as np
import pandas as pd
import scipy.stats
from statsmodels.stats.multitest import multipletests
from .table_check import table_check


def univariate_2class(DataTable, PeakTable, group, posclass, parametric=True):
    """Creates a table of univariate statistics (2 class)."""
    
    # Error checks
    table_check(DataTable, PeakTable, print_statement=False)
    if group not in DataTable:
        raise ValueError("Column '{}' does not exist in DataTable".format(group))
    if posclass not in DataTable[group].unique():
        raise ValueError("Positive class was not found in '{}' column.".format(group))
    if len(DataTable[group].unique()) != 2:
        raise ValueError("Column '{}' should have exactly 2 groups".format(group))
    
    # Get x0, x1
    peaklist = PeakTable['Name']
    x = DataTable[peaklist]
    y = DataTable[group]
    x1 = x[y == posclass] #x1 refers to posclass
    x0 = x[y != posclass]
    
    # Create stats table (include Idx, Name and Label)
    stats = pd.DataFrame()
    stats['Idx'] = PeakTable['Idx']
    stats['Name'] = PeakTable['Name']
    stats['Label'] = PeakTable['Label']
    
    if parametric is True:
        #Calculate mean and std
        stats['Mean0'] = np.nanmean(x0, axis=0)
        Mean095CI = (1.96 * (np.nanstd(x0, ddof=1, axis=0) / np.sqrt(x0.notnull().sum()))).values
        stats['Mean0-95CI'] = list(zip(np.round(stats['Mean0'] - Mean095CI,2), np.round(stats['Mean0'] + Mean095CI,2)))
        stats['Mean1'] = np.nanmean(x1, axis=0)
        Mean195CI = (1.96 * (np.nanstd(x1, ddof=1, axis=0) / np.sqrt(x1.notnull().sum()))).values
        stats['Mean1-95CI'] = list(zip(np.round(stats['Mean1'] - Mean195CI,2), np.round(stats['Mean1'] + Mean195CI,2)))
        
        #Sign
        sign = (np.nanmedian(x1, axis=0) / np.nanmedian(x0, axis=0))
        sign = np.where(sign > 1, 1, 0)
        stats['Sign'] = sign
        
        #T test 
        xnp = x.values 
        t = scipy.stats.ttest_ind(x0, x1, nan_policy='omit')
        stats['TTestStat'] = t[0]
        stats['TTestPvalue'] = t[1]
        
        # BH correction 
        bonTTest = multipletests(t[1], method='fdr_bh') 
        # Round BH correction p-value
        stats['QTTest'] = bonTTest[1]
    else:
        #Calculate median
        stats['Median0'] = np.nanmedian(x0, axis=0)
        x0_boot = []
        for i in range(100):
            x0_copy = x0.copy()
            x0_copy = x0_copy.iloc[np.random.randint(0, len(x0_copy), size=len(x0_copy))] 
            x0_boot.append(np.nanmedian(x0_copy, axis=0))
        x0low = np.round(np.percentile(x0_boot, 2.5, axis=0),2)
        x0high = np.round(np.percentile(x0_boot, 97.5, axis=0),2)
        stats['Median0-95CI'] = list(zip(x0low,x0high))
        
        stats['Median1'] = np.nanmedian(x1, axis=0)
        x1_boot = []
        for i in range(100):
            x1_copy = x1.copy()
            x1_copy = x1_copy.iloc[np.random.randint(0, len(x1_copy), size=len(x1_copy))] 
            x1_boot.append(np.nanmedian(x1_copy, axis=0))
        x1low = np.round(np.percentile(x1_boot, 2.5, axis=0),2)
        x1high = np.round(np.percentile(x1_boot, 97.5, axis=0),2)
        stats['Median1-95CI'] = list(zip(x1low,x1high))
        stats['MedianFC'] = np.nanmedian(x1, axis=0) / np.nanmedian(x0, axis=0)
        
        #Sign
        sign = (np.nanmedian(x1, axis=0) / np.nanmedian(x0, axis=0))
        sign = np.where(sign > 1, 1, 0)
        stats['Sign'] = sign
        
        #Man-Whitney U
        m = []
        for i in range(len(x.T)):
            manwstat = scipy.stats.mannwhitneyu(x0.values[:,i], x1.values[:,i])
            m.append(manwstat)
        stats['MannWhitneyU'] = [i[0] for i in m] 
        stats['MannWhitneyPvalue'] = [i[1] for i in m] 
        
        # BH correction 
        bonMannWhitney = multipletests(stats["MannWhitneyPvalue"], method='fdr_bh') 
        stats['QMannWhitney'] = bonMannWhitney[1]
    
    #Calculate total missing and total missing %
    nannum = x.isnull().sum() 
    nanperc = nannum / x.shape[0]
    stats['TotalMissing'] = nannum.values
    stats['PercTotalMissing'] = np.round(nanperc.values * 100, 3)
    
    #Calculating missing % for group 0, and group 1... 
    nanperc_0 = x0.isnull().sum() / len(x0)
    nanperc_1 = x1.isnull().sum() / len(x1)
    stats['Perc0Missing'] = np.round(nanperc_0.values * 100, 3)
    stats['Perc1Missing'] = np.round(nanperc_1.values * 100, 3)
    
    #Shapiro-Wilk     
    s = []
    for i in range(len(x.T)):
        newx = x.values[:,i][~np.isnan(x.values[:,i])]
        shapstat = scipy.stats.shapiro(newx)
        s.append(shapstat)
    stats['ShapiroW'] = [i[0] for i in s] #Shapstat is a list of tuples, this is an approach to get this out of the tuple
    stats['ShapiroPvalue'] = [i[1] for i in s]
    
    #Levenes       
    l = []
    for i in range(len(x.T)):
        newx0 = x0.values[:,i][~np.isnan(x0.values[:,i])]
        newx1 = x1.values[:,i][~np.isnan(x1.values[:,i])]
        levstat = scipy.stats.levene(newx0, newx1)
        l.append(levstat)         
    stats['LeveneW'] = [i[0] for i in l] 
    stats['LevenePvalue'] = [i[1] for i in l]    
    return stats



import numpy as np
import pandas as pd
from sklearn import decomposition
from bokeh.plotting import output_notebook, show 
from bokeh.layouts import gridplot
from .scatter import scatter
from ..utils import ci95_ellipse


def pca(X, pcx=1, pcy=2, group_label=None, sample_label=None, peak_label=None):
    """Creates a PCA scores and loadings plot using Bokeh."""
    
    if len(group_label) != len(X):
        raise V
    
    # Set model
    model = decomposition.PCA() 
    model.fit(X)
    scores_ = model.transform(X)
    explained_var_ = model.explained_variance_ratio_ * 100
    
    # Extract scores, explained variance, and loadings for pcx and pcy
    x_score = scores_[:,(pcx-1)]
    y_score = scores_[:,(pcy-1)]
    x_expvar = explained_var_[(pcx-1)]
    y_expvar = explained_var_[(pcy-1)]
    x_load = model.components_[:,(pcx-1)]
    y_load = model.components_[:,(pcy-1)]
        
    # Scores plot
    fig_score = scatter(x_score, y_score, group=group_label, label=sample_label, size=5, xlabel='PC {} ({:0.1f}%)'.format(pcx, x_expvar), ylabel='PC {} ({:0.1f}%)'.format(pcy, y_expvar), title='PCA Score Plot (PC{} vs. PC{})'.format(pcx, pcy), font_size='15pt', width=490, height=430, hover_xy=False)
    
    # Loadings plot
    fig_load = scatter(x_load, y_load, size=7, label=peak_label, xlabel='PC {} ({:0.1f}%)'.format(pcx, x_expvar), ylabel='PC {} ({:0.1f}%)'.format(pcy, y_expvar), title='PCA Loadings Plot (PC{} vs. PC{})'.format(pcx, pcy), font_size='15pt', width=490, height=430, hover_xy=False, shape='triangle', legend=False)
    
    # Score plot extra: 95% confidence ellipse using PCA 
    unique_group = np.sort(np.unique(group_label))

    # Set colour per group
    list_color = ['red', 'blue', 'green', 'black']
    while len(list_color) < len(unique_group): # Loop over list_color if number of groups > len(list_colour)
        list_color += list_color 

    # Add 95% confidence ellipse for each unique group in a loop
    for i in range(len(unique_group)):
        # Get scores for the corresponding group
        group_i_x = []
        group_i_y = []
        for j in range(len(group_label)):
            if group_label[j] == unique_group[i]:
                group_i_x.append(x_score[j])
                group_i_y.append(y_score[j])
                
        # Calculate ci95 ellipse for each group
        data_circ_group = pd.DataFrame({"0": group_i_x, "1": group_i_y})
        m, outside_m = ci95_ellipse(data_circ_group, type="mean")
        p, outside_p = ci95_ellipse(data_circ_group, type="pop")
        
        # Plot ci95 ellipse outer line
        fig_score.line(m[:,0], m[:,1], color=list_color[i], line_width=2, alpha=0.8, line_dash="solid")
        fig_score.line(p[:,0], p[:,1], color=list_color[i], alpha=0.4)
        
        # Plot ci95 ellipse shade
        fig_score.patch(m[:,0], m[:,1], color=list_color[i], alpha=0.07)
        fig_score.patch(p[:,0], p[:,1], color=list_color[i], alpha=0.01)
        
        
    
    # Output this figure with fig_score and fig_load
    output_notebook()    
    fig = gridplot([[fig_score, fig_load]])
    show(fig)

        



import pandas as pd
import numpy as np
from os import path
from .table_check import table_check


def load_dataXL(filename, DataSheet, PeakSheet):
    """Loads and validates the DataFile and PeakFile."""

    if path.isfile(filename) is False:
        raise ValueError("{} does not exist.".format(filename))

    if not filename.endswith('.xlsx'):
        raise ValueError("{} should be a .xlsx file.".format(filename))

    # LOAD PEAK DATA
    print("Loadings PeakFile: {}".format(PeakSheet))
    PeakTable = pd.read_excel(filename, sheet_name=PeakSheet)

    # LOAD DATA TABLE
    print("Loadings DataFile: {}".format(DataSheet))
    DataTable = pd.read_excel(filename, sheet_name=DataSheet)

    # Replace with nans
    DataTable = DataTable.replace(-99, np.nan)
    DataTable = DataTable.replace('.', np.nan)
    DataTable = DataTable.replace(' ', np.nan)
    
    # Error checks
    table_check(DataTable, PeakTable, print_statement=True)

    print("TOTAL SAMPLES: {} TOTAL PEAKS: {}".format(len(DataTable), len(PeakTable)))
    print("Done!")

    return DataTable, PeakTable



import numpy as np
from bokeh.models import Band, HoverTool 
from bokeh.plotting import ColumnDataSource, figure
from scipy import interp
from sklearn import metrics
from sklearn.utils import resample


def roc_plot(fpr, tpr, tpr_ci, width=450, height=350, xlabel="1-Specificity", ylabel="Sensitivity", legend=True, label_font_size="13pt", title="", errorbar=False):
    """Creates a rocplot using Bokeh."""
    
    # Get CI
    tpr_lowci = tpr_ci[0]
    tpr_uppci = tpr_ci[1]
    auc = metrics.auc(fpr, tpr)
    
    # specificity and ci-interval for HoverTool
    spec = 1 - fpr
    ci = (tpr_uppci - tpr_lowci) / 2

    # Figure
    data = {"x": fpr, "y": tpr, "lowci": tpr_lowci, "uppci": tpr_uppci, "spec":spec, "ci": ci}
    source = ColumnDataSource(data=data)
    fig = figure(title=title, plot_width=width, plot_height=height, x_axis_label=xlabel, y_axis_label=ylabel, x_range=(-0.06, 1.06), y_range=(-0.06, 1.06))
    
    # Figure: add line
    fig.line([0, 1], [0, 1], color="black", line_dash="dashed", line_width=2.5, legend="Equal distribution line")
    figline = fig.line("x", "y", color="green", line_width=3.5, alpha=0.6, legend="ROC Curve (Train)", source=source)
    fig.add_tools(HoverTool(renderers=[figline],tooltips=[("Specificity", "@spec{1.111}"),("Sensitivity", "@y{1.111} (+/- @ci{1.111})"),],))

    # Figure: add 95CI band
    figband = Band(base="x", lower="lowci", upper="uppci", level="underlay", fill_alpha=0.1, line_width=1, line_color="black", fill_color="green", source=source)
    fig.add_layout(figband)
    
    # Figure: add errorbar  spec =  1 - fpr
    if errorbar is not False:
        idx = np.abs(fpr - (1 - errorbar)).argmin() # this find the closest value in fpr to errorbar fpr
        fpr_eb = fpr[idx] 
        tpr_eb = tpr[idx]
        tpr_lowci_eb = tpr_lowci[idx]
        tpr_uppci_eb = tpr_uppci[idx]
        roc_whisker_line = fig.multi_line([[fpr_eb, fpr_eb]], [[tpr_lowci_eb, tpr_uppci_eb]], line_alpha=1, line_color="black")
        roc_whisker_bot = fig.multi_line([[fpr_eb - 0.03, fpr_eb + 0.03]], [[tpr_lowci_eb, tpr_lowci_eb]], line_color="black")
        roc_whisker_top = fig.multi_line([[fpr_eb - 0.03, fpr_eb + 0.03]], [[tpr_uppci_eb, tpr_uppci_eb]], line_alpha=1, line_color="black")
        fig.circle([fpr_eb], [tpr_eb], size=8, fill_alpha=1, line_alpha=1, line_color="black", fill_color="white",)  

    # Change font size
    fig.title.text_font_size = "11pt"
    fig.xaxis.axis_label_text_font_size = label_font_size
    fig.yaxis.axis_label_text_font_size = label_font_size
    fig.legend.label_text_font = "10pt"

    # Extra padding
    fig.min_border_left = 20
    fig.min_border_right = 20
    fig.min_border_top = 20
    fig.min_border_bottom = 20

    # Edit legend
    fig.legend.location = "bottom_right"
    fig.legend.label_text_font_size = "10pt"
    if legend is False:
        fig.legend.visible = False
    return fig




import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold
from bokeh.layouts import gridplot 
from bokeh.plotting import figure
from bokeh.models import ColumnDataSource
from bokeh.models import Circle, HoverTool, TapTool
from tqdm import tqdm
from bokeh.plotting import output_notebook, show 
from .BaseCrossVal import BaseCrossVal
from ..utils import binary_metrics


class kfold(BaseCrossVal):
    """ Exhaustitive search over param_dict calculating binary metrics. 
    
    Parameters
    ----------
    model : object
        This object is assumed to store bootlist attributes in .model (e.g. modelPLS.model.x_scores_).
    
    X : array-like, shape = [n_samples, n_features]
        Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.
    
    Y : array-like, shape = [n_samples, 1]
        Response variables, where n_samples is the number of samples.
        
    param_dict : dict
        List of attributes to calculate and return bootstrap confidence intervals.
    
    folds: : a positive integer, (default 10)
        The number of folds used in the computation.

    
    bootnum : a positive integer, (default 100)
        The number of bootstrap samples used in the computation for the plot. 
 

    Attributes
    -------
    Run: Runs all functions prior to plot.
    
    Plot: Creates a R2/Q2 plot.
    
  
    """
    
    def __init__(self, model, X, Y, param_dict, folds=10, bootnum=100):
        super().__init__(model=model, X=X, Y=Y, param_dict=param_dict, folds=folds, bootnum=bootnum)
        self.crossval_idx = StratifiedKFold(n_splits=folds)
        
    def calc_ypred(self):
        """Calculates ypred full and ypred cv."""
        self.ypred_full = []
        self.ypred_cv = []
        for params in self.param_list:
            # Set hyper-parameters
            params_i = params
            model_i = self.model(**params_i)
            # Full
            model_i.train(self.X, self.Y)
            ypred_full_i = model_i.test(self.X)
            self.ypred_full.append(ypred_full_i)
            # CV (for each fold)
            ypred_cv_i = self._calc_cv_ypred(model_i, self.X, self.Y)
            self.ypred_cv.append(ypred_cv_i)

    def calc_stats(self):
        """Calculates binary statistics from ypred full and ypred cv."""
        stats_list = []
        self.full_boot_metrics = []
        self.cv_boot_metrics = []
        for i in range(len(self.param_list)):
            # Create dictionaries with binary_metrics
            stats_full_i = binary_metrics(self.Y, self.ypred_full[i])
            stats_cv_i = binary_metrics(self.Y, self.ypred_cv[i])
            # Add _cv to all metrics in stats_cv and change R2_cv to Q2
            stats_cv_i = {k+'_cv': v for k, v in stats_cv_i.items()}
            stats_cv_i["Q2"] = stats_cv_i.pop("R2_cv")
            # Combine and append
            stats_combined = {**stats_full_i, **stats_cv_i}
            stats_list.append(stats_combined)
        self.table = self._format_table(stats_list) #Transpose, Add headers
        return self.table
    
    def run(self):
        """Runs all functions prior to plot."""
        self.calc_ypred()
        self.calc_stats()
        if self.bootnum > 1:
            self.calc_ypred_boot()
            self.calc_stats_boot()
            
    def calc_ypred_boot(self):
        """Calculates ypred full and ypred cv for each bootstrap resample."""
        self.ytrue_boot = []
        self.ypred_full_boot = []
        self.ypred_cv_boot = []
        for i in tqdm(range(self.bootnum), desc= "Kfold"):
            bootidx_i = np.random.choice(len(self.Y), len(self.Y))
            newX = self.X[bootidx_i, :]
            newY = self.Y[bootidx_i]
            ypred_full_nboot_i = []
            ypred_cv_nboot_i = []
            for params in self.param_list:
                # Set hyper-parameters
                model_i = self.model(**params)
                # Full
                model_i.train(newX, newY)
                ypred_full_i = model_i.test(newX)
                ypred_full_nboot_i.append(ypred_full_i)
                # cv
                ypred_cv_i = self._calc_cv_ypred(model_i, newX, newY)
                ypred_cv_nboot_i.append(ypred_cv_i)
            self.ytrue_boot.append(newY)
            self.ypred_full_boot.append(ypred_full_nboot_i)
            self.ypred_cv_boot.append(ypred_cv_nboot_i)
            
    def calc_stats_boot(self):
        """Calculates binary statistics from ypred full and ypred cv for each bootstrap resample."""
        self.full_boot_metrics = []
        self.cv_boot_metrics = []
        for i in range(len(self.param_list)):
            stats_full_i = []
            stats_cv_i = []
            for j in range(self.bootnum):
                stats_full = binary_metrics(self.ytrue_boot[j], self.ypred_full_boot[j][i])
                stats_full_i.append(stats_full)
                stats_cv = binary_metrics(self.ytrue_boot[j], self.ypred_cv_boot[j][i])
                stats_cv_i.append(stats_cv)
            self.full_boot_metrics.append(stats_full_i)
            self.cv_boot_metrics.append(stats_cv_i)
            
    def _calc_cv_ypred(self, model_i, X, Y):
        """Method used to calculate ypred cv."""
        ypred_cv_i = [None] * len(Y)
        for train, test in self.crossval_idx.split(self.X, self.Y):
            X_train = X[train, :]
            Y_train = Y[train]
            X_test = X[test, :]
            Y_test = Y[test]
            model_i.train(X_train, Y_train)
            ypred_cv_i_j = model_i.test(X_test, Y_test)
            # Return value to y_pred_cv in the correct position # Better way to do this
            for (idx, val) in zip(test, ypred_cv_i_j):
                ypred_cv_i[idx] = val.tolist()
        return ypred_cv_i
    
    def _format_table(self, stats_list):
        """Make stats pretty (pandas table -> proper names in columns)."""
        table = pd.DataFrame(stats_list).T
        param_list_string = []
        for i in range(len(self.param_list)):
            param_list_string.append(str(self.param_list[i]))
        table.columns = param_list_string
        return table

    def plot(self):
        "Create a R2/Q2 plot."
        
        # get r2, q2, and diff
        r2 = self.table.loc["R2"]
        q2 = self.table.loc["Q2"]
        diff = r2 - q2
        
        # round r2, q2, and diff for hovertool
        r2_text = []
        q2_text = []
        diff_text = []
        for j in range(len(r2)):
            r2_text.append("%.2f" % round(r2[j], 2))
            q2_text.append("%.2f" % round(q2[j], 2))
            diff_text.append("%.2f" % round(diff[j], 2))
        
        # get key, values (as string) from param_dict (key -> title, values -> x axis values)
        for k, v in self.param_dict.items():
            key = k
            values = v
        values_string = [str(i) for i in values]
        
        # store data in ColumnDataSource for Bokeh
        data = dict(r2=r2, q2=q2, diff=diff, r2_text=r2_text, q2_text=q2_text, diff_text=diff_text, values_string=values_string)
        source = ColumnDataSource(data=data)
        
        fig1_yrange = (min(diff) - max(0.1 * (min(diff)), 0.03), max(diff) + max(0.1 * (max(diff)), 0.03))
        fig1_xrange = (min(q2) - max(0.1 * (min(q2)), 0.03), max(q2) + max(0.1 * (max(q2)), 0.03))
        
        # Figure 1 (DIFFERENCE (R2 - Q2) vs. Q2)
        fig1 = figure(x_axis_label="Q2", y_axis_label="DIFFERENCE (R2 - Q2)", title="DIFFERENCE (R2 - Q2) vs. Q2", tools="tap,pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select", y_range=fig1_yrange, x_range=fig1_xrange,plot_width=480, plot_height=400)
        fig1.title.text_font_size = "14pt"
        
        # Figure 1: Add a line
        fig1_line = fig1.line(q2, diff, line_width=2, line_color="black", line_alpha=0.25)
        
        # Figure 1: Add circles (interactive click)
        fig1_circ = fig1.circle("q2", "diff", size=10, alpha=0.7, color="green", source=source)
        fig1_circ.selection_glyph = Circle(fill_color="green", line_width=2, line_color="black")
        fig1_circ.nonselection_glyph.fill_color = "green"
        fig1_circ.nonselection_glyph.fill_alpha = 0.4
        fig1_circ.nonselection_glyph.line_color = "white"
        
        # Figure 1: Add hovertool
        fig1.add_tools(HoverTool(renderers=[fig1_circ],tooltips=[(key, "@values_string"), ("R2", "@r2_text"), ("Q2", "@q2_text"), ("Diff", "@diff_text"),],))
        
        # Figure 1: Extra formating
        fig1.axis.major_label_text_font_size = "8pt"
        fig1.xaxis.axis_label_text_font_size = "12pt"
        fig1.yaxis.axis_label_text_font_size = "12pt"

        # Figure 2: R2/Q2 
        fig2 = figure(x_axis_label=key, y_axis_label="Value", title="R2/Q2 vs. {}".format(key), plot_width=480, plot_height=400, x_range=pd.unique(values_string), y_range=(0, 1.1), tools="pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select")
        fig2.title.text_font_size = "14pt"
        
        # Figure 2: add confidence intervals if bootnum > 1
        if self.bootnum > 1:
            lower_ci_r2 = []
            upper_ci_r2 = []
            lower_ci_q2 = []
            upper_ci_q2 = []
            # Get all upper, lower 95% CI (r2/q2) for each specific n_component and append 
            for m in range(len(self.full_boot_metrics)):
                r2_boot = []
                q2_boot = []
                for k in range(len(self.full_boot_metrics[0])):
                    r2_boot.append(self.full_boot_metrics[m][k]['R2'])
                    q2_boot.append(self.cv_boot_metrics[m][k]['R2']) # Wasn't renamed to Q2
                # Calculated percentile 95% CI and append
                r2_bias = np.mean(r2_boot) - r2[m]
                q2_bias = np.mean(q2_boot) - q2[m]
                lower_ci_r2.append(np.percentile(r2_boot, 2.5) - r2_bias)
                upper_ci_r2.append(np.percentile(r2_boot, 97.5) - r2_bias)
                lower_ci_q2.append(np.percentile(q2_boot, 2.5) - q2_bias)
                upper_ci_q2.append(np.percentile(q2_boot, 97.5) - q2_bias)
            
            # Plot as a patch
            x_patch = np.hstack((values_string, values_string[::-1]))
            y_patch_r2 = np.hstack((lower_ci_r2, upper_ci_r2[::-1]))
            fig2.patch(x_patch, y_patch_r2, alpha=0.10, color="red")
            y_patch_q2 = np.hstack((lower_ci_q2, upper_ci_q2[::-1]))
            fig2.patch(x_patch, y_patch_q2, alpha=0.10, color="blue")
        
        # Figure 2: add r2
        fig2_line_r2 = fig2.line(values_string, r2, line_color="red", line_width=2)
        fig2_circ_r2 = fig2.circle("values_string", "r2", line_color="red", fill_color="white", fill_alpha=1, size=8, source=source)
        fig2_circ_r2.selection_glyph = Circle(line_color="red", fill_color="white", line_width=2)
        fig2_circ_r2.nonselection_glyph.line_color = "red"
        fig2_circ_r2.nonselection_glyph.fill_color = "white"
        fig2_circ_r2.nonselection_glyph.line_alpha = 0.4
        
        # Figure 2: add q2
        fig2_line_q2 = fig2.line(values_string, q2, line_color="blue", line_width=2)
        fig2_circ_q2 = fig2.circle("values_string", "q2", line_color="blue", fill_color="white", fill_alpha=1, size=8, source=source)
        fig2_circ_q2.selection_glyph = Circle(line_color="blue", fill_color="white", line_width=2)
        fig2_circ_q2.nonselection_glyph.line_color = "blue"
        fig2_circ_q2.nonselection_glyph.fill_color = "white"
        fig2_circ_q2.nonselection_glyph.line_alpha = 0.4
        
        # Add hovertool and taptool
        fig2.add_tools(HoverTool(renderers=[fig2_circ_r2], tooltips=[("R2", "@r2")], mode="vline"))
        fig2.add_tools(HoverTool(renderers=[fig2_circ_q2], tooltips=[("Q2", "@q2")], mode="vline"))
        fig2.add_tools(TapTool(renderers=[fig2_circ_r2, fig2_circ_q2]))

        # Create a grid and output figures
        grid = np.full((1, 2), None) 
        grid[0, 0] = fig1
        grid[0, 1] = fig2
        fig = gridplot(grid.tolist(), merge_tools=True)
        output_notebook()
        show(fig)
    




import numpy as np


def scale(x, axis=0, ddof=1, method='auto', return_mu_sigma=False, mu='default', sigma='default'):
    """Scales x (which can include nans) with method: 'auto', 'pareto', 'vast', 'range', or 'level'."""
    
    x = np.asanyarray(x)
    
    # Calculate mu and sigma if set to 'default'
    if mu is 'default':
        mu = np.nanmean(x, axis=axis)
    if sigma is 'default':
        sigma = np.nanstd(x, axis=axis, ddof=ddof)
        sigma = np.where(sigma > 0, sigma, 1) # if a value in sigma equals 0 it is converted to 1
    
    # expands the shape of the array when required
    if axis and mu.ndim < x.ndim:
        mu = np.expand_dims(mu, axis=axis)
        sigma = np.expand_dims(sigma, axis=axis)
        
    # Error check before scaling
    if len(mu) != len(x.T):
        raise ValueError("Length of mu array does not match x matrix.")
    if len(sigma) != len(x.T):
        raise ValueError("Length of sigma array does not match x matrix.")

    # Scale based on selected method
    if method is "auto":
        z = (x - mu) / sigma
    elif method is "pareto":
        z = (x - mu) / np.sqrt(sigma)
    elif method is "vast":
        z = ((x - mu) / sigma) * (mu / sigma)
    elif method is "level":
        z = (x - mu) / mu
    else:
        raise ValueError("Method has to be either 'auto', 'pareto', 'vast', or 'level'.")
    
    if return_mu_sigma is True: 
        return mu, sigma
    else:
        return z



import numpy as np
from sklearn import metrics 
from sklearn.metrics import confusion_matrix, roc_auc_score
from sklearn.utils import resample
from .binary_metrics import binary_metrics


def roc_calculate(Ytrue, Yscore, bootnum=1000, metric=None, val=None):
    """Calculates required metrics for the roc plot function."""
    
    # Get fpr, tpr
    fpr, tpr, threshold = metrics.roc_curve(Ytrue, Yscore, pos_label=1, drop_intermediate=False)
    
    # fpr, tpr with drop_intermediates for fpr = 0 (useful for plot... since we plot specificity on x-axis, we don't need intermediates when fpr=0)
    tpr0 = tpr[fpr == 0][-1]
    tpr = np.concatenate([[tpr0], tpr[fpr > 0]])
    fpr = np.concatenate([[0], fpr[fpr > 0]])
    
    # if metric is provided, calculate stats
    if metric is not None:
        specificity, sensitivity, threshold = get_spec_sens_cuttoff(Ytrue, Yscore, metric, val)
        stats = get_stats(Ytrue, Yscore, specificity)
        stats['val_specificity'] = specificity
        stats['val_sensitivity'] = specificity
        stats['val_cutoffscore'] = threshold 
        
    # bootstrap using vertical averaging
    tpr_boot = []
    boot_stats = []
    for i in range(bootnum):
        # Resample and get tpr, fpr
        Ytrue_res, Yscore_res = resample(Ytrue, Yscore) 
        fpr_res, tpr_res, threshold_res = metrics.roc_curve(Ytrue_res, Yscore_res, pos_label=1, drop_intermediate=False)

        # Drop intermediates when fpr=0
        tpr0_res = tpr_res[fpr_res == 0][-1]
        tpr_res = np.concatenate([[tpr0_res], tpr_res[fpr_res > 0]])
        fpr_res = np.concatenate([[0], fpr_res[fpr_res > 0]])

        # Vertical averaging... use closest fpr_res to fpr, and append the corresponding tpr
        idx = [np.abs(i - fpr_res).argmin() for i in fpr]
        tpr_list = tpr_res[idx]
        tpr_boot.append(tpr_list)
        
        # if metric is provided, calculate stats
        if metric is not None:
            stats_res = get_stats(Ytrue_res, Yscore_res, specificity)
            boot_stats.append(stats_res)
            
    # Get CI for bootstat
    if metric is not None:
        bootci_stats = {}
        for i in boot_stats[0].keys():
            stats_i = [k[i] for k in boot_stats]
            lowci = np.percentile(stats_i,2.5)
            uppci = np.percentile(stats_i, 97.5)
            bootci_stats[i] = [lowci, uppci]
        
    # Get CI for tpr 
    tpr_lowci = np.percentile(tpr_boot, 2.5, axis=0)
    tpr_uppci = np.percentile(tpr_boot, 97.5, axis=0)

    # Add the starting 0 
    tpr = np.insert(tpr, 0, 0)
    fpr = np.insert(fpr, 0, 0)
    tpr_lowci = np.insert(tpr_lowci, 0, 0)
    tpr_uppci = np.insert(tpr_uppci, 0, 0)
    
    # Concatenate tpr_ci
    tpr_ci = np.array([tpr_lowci, tpr_uppci])
    
    if metric is None:
        return fpr, tpr, tpr_ci
    else:
        return fpr, tpr, tpr_ci, stats, bootci_stats
    

def get_sens_spec(Ytrue, Yscore, cuttoff_val):
    """Get sensitivity and specificity from cutoff value."""
    Yscore_round = np.where(np.array(Yscore) > cuttoff_val, 1, 0) 
    tn, fp, fn, tp = metrics.confusion_matrix(Ytrue, Yscore_round).ravel()
    sensitivity  = tp / (tp + fn)
    specificity  = tn / (tn + fp)
    return sensitivity, specificity

def get_sens_cuttoff(Ytrue, Yscore, specificity_val):
    """Get sensitivity and cuttoff value from specificity."""
    fpr0 = 1 - specificity_val
    fpr, sensitivity , thresholds = metrics.roc_curve(Ytrue, Yscore, pos_label=1, drop_intermediate=False)
    idx = np.abs(fpr - fpr0).argmin() # this find the closest value in fpr to fpr0
    return sensitivity[idx], thresholds[idx]

def get_spec_sens_cuttoff(Ytrue, Yscore, metric, val):
    """Return specificity, sensitivity, cutoff value provided the metric and value used."""
    if metric == 'specificity':
        specificity = val
        sensitivity, threshold = get_sens_cuttoff(Ytrue, Yscore, val)
    elif metric == 'cutoffscore':
        threshold = val
        sensitivity, specificity = get_sens_spec(Ytrue, Yscore, val)
    return specificity, sensitivity, threshold

def get_stats(Ytrue, Yscore, specificity):
    """Calculates binary metrics given the specificity."""
    sensitivity, cutoffscore = get_sens_cuttoff(Ytrue, Yscore, specificity)
    stats = binary_metrics(Ytrue, Yscore, cut_off=cutoffscore)
    return stats


from .binary_metrics import binary_metrics
from .ci95_ellipse import ci95_ellipse
from .knnimpute import knnimpute
from .load_dataXL import load_dataXL
from .roc_calculate import roc_calculate
from .scale import scale
from .nested_getattr import nested_getattr
from .table_check import table_check
from .univariate_2class import univariate_2class

__all__ = [
    "binary_metrics",
    "ci95_ellipse",
    "knnimpute",
    "load_dataXL",
    "scale",
    "nested_getattr",
    "table_check",
    "univariate_2class",
]



import numpy as np
import warnings
from sklearn.metrics import confusion_matrix, roc_auc_score


def binary_metrics(Y_true, Y_pred, cut_off=0.5):
    """Return a dict of binary stats with the following metrics: R2, auc, accuracy, precision, sensitivity, specificity, and F1 score."""
    
    Y_pred_round = np.where(np.array(Y_pred) >= cut_off, 1, 0) 
    tn, fp, fn, tp = confusion_matrix(Y_true, Y_pred_round).ravel() 
    
    # Binary statistics dictionary
    stats = {}
    stats["R2"] = 1 - (sum((Y_true - Y_pred) ** 2) / sum((Y_true - np.mean(Y_true)) ** 2))
    stats["auc"] = roc_auc_score(Y_true, Y_pred)
    stats["accuracy"] = safe_div((tp + tn) , (tp + tn + fp + fn))
    stats["precision"] = safe_div((tp) , (tp + fp))
    stats["sensitivity"] = safe_div((tp) , (tp + fn))
    stats["specificity"] = safe_div((tn) , (tn + fp))
    stats["F1score"] = safe_div((2 * tp) , (2 * tp + fp + fn))

    return stats

def safe_div(x,y):
    if y == 0:
        return np.nan
    return x / y


from abc import ABC, abstractmethod, abstractproperty
import numpy as np
import pandas as pd
import scipy
from bokeh.layouts import widgetbox, gridplot, column, row, layout
from bokeh.models import HoverTool, Band
from bokeh.models.widgets import DataTable, Div, TableColumn
from bokeh.models.annotations import Title
from bokeh.plotting import ColumnDataSource, figure, output_notebook, show 
from scipy import interp
from sklearn import metrics
from sklearn.utils import resample
from ..bootstrap import Perc, BC, BCA
from ..plot import scatter, scatterCI, boxplot, distribution, permutation_test, roc_plot
from ..utils import binary_metrics, roc_calculate


class BaseModel(ABC):
    """Base class for models: PLS_SIMPLS."""

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def train(self):
        """Trains the model."""
        pass

    @abstractmethod
    def test(self):
        """Tests the model."""
        pass

    @abstractproperty
    def bootlist(self):
        """A list of attributes for bootstrap resampling."""
        pass

    def evaluate(self, testset=None, specificity=False, cutoffscore=False, bootnum=1000):
        """Returns a figure containing a Violin plot, Distribution plot, ROC plot and Binary Metrics statistics."""
        
        Ytrue_train = self.Y
        Yscore_train = self.Y_pred.flatten()
        
        # Get Ytrue_test, Yscore_test from testset
        if testset is not None:
            Ytrue_test = np.array(testset[0])
            Yscore_test = np.array(testset[1])
            
            # Error checking
            if len(Ytrue_test) != len(Yscore_test):
                raise ValueError("evaluate can't be used as length of Ytrue does not match length of Yscore in test set.")
            if len(np.unique(Ytrue_test)) != 2:
                raise ValueError("Ytrue_test needs to have 2 groups. There is {}".format(len(np.unique(Y))))
            if np.sort(np.unique(Ytrue_test))[0] != 0:
                raise ValueError("Ytrue_test should only contain 0s and 1s.")
            if np.sort(np.unique(Ytrue_test))[1] != 1:
                raise ValueError("Ytrue_test should only contain 0s and 1s.")

            # Get Yscore_combined and Ytrue_combined_name (Labeled Ytrue)
            Yscore_combined = np.concatenate([Yscore_train, Yscore_test])
            Ytrue_combined = np.concatenate([Ytrue_train, Ytrue_test + 2]) # Each Ytrue per group is unique
            Ytrue_combined_name = Ytrue_combined.astype(np.str)
            Ytrue_combined_name[Ytrue_combined == 0] = "Train (0)"
            Ytrue_combined_name[Ytrue_combined == 1] = "Train (1)"
            Ytrue_combined_name[Ytrue_combined == 2] = "Test (0)"
            Ytrue_combined_name[Ytrue_combined == 3] = "Test (1)"
        
        # Expliclity states which metric and value is used for the error_bar
        if specificity is not False:
            metric = 'specificity'
            val = specificity
        elif cutoffscore is not False:
            metric = 'cutoffscore'
            val = cutoffscore
        else:
            metric = 'specificity'
            val = 0.8
            
        # ROC plot
        tpr, fpr, tpr_ci, stats, stats_bootci = roc_calculate(Ytrue_train, Yscore_train, bootnum=100, metric=metric, val=val)
        roc_title = "Specificity: {}".format(np.round(stats['val_specificity'],2))
        roc_bokeh = roc_plot(tpr, fpr, tpr_ci, width=320, height=315, title=roc_title, errorbar=stats['val_specificity'])
        if testset is not None:
            fpr_test, tpr_test, threshold_test = metrics.roc_curve(Ytrue_test, Yscore_test, pos_label=1, drop_intermediate=False)
            fpr_test = np.insert(fpr_test, 0, 0)
            tpr_test = np.insert(tpr_test, 0, 0)
            roc_bokeh.line(fpr_test, tpr_test, color="red", line_width=3.5, alpha=0.6, legend="ROC Curve (Test)") # Add ROC Curve(Test) to roc_bokeh

        # Violin plot
        if testset is None:
            violin_bokeh = boxplot(Yscore_train, Ytrue_train, title="", xlabel="Class", ylabel="Predicted Score", violin=True, color=["#FFCCCC", "#CCE5FF"], width=320, height=315)
        else:
            violin_bokeh = boxplot(Yscore_combined, Ytrue_combined_name, title="", xlabel="Class", ylabel="Predicted Score", violin=True, color=["#fcaeae", "#aed3f9","#FFCCCC", "#CCE5FF"], width=320, height=315, group_name=['Train (0)', 'Test (0)', 'Train (1)', 'Test (1)'], group_name_sort=['Test (0)', 'Test (1)', 'Train (0)', 'Train (1)'])
        violin_bokeh.multi_line([[-100, 100]], [[stats['val_cutoffscore'], stats['val_cutoffscore']]], line_color="black", line_width=2, line_alpha=1.0, line_dash="dashed")

        # Distribution plot
        if testset is None:
            dist_bokeh = distribution(Yscore_train, group=Ytrue_train, kde=True, title="", xlabel="Predicted Score", ylabel="p.d.f.", width=320, height=315)
        else:
            dist_bokeh = distribution(Yscore_combined, group=Ytrue_combined_name, kde=True, title="", xlabel="Predicted Score", ylabel="p.d.f.", width=320, height=315)
        dist_bokeh.multi_line([[stats['val_cutoffscore'], stats['val_cutoffscore']]], [[-100, 100]], line_color="black", line_width=2, line_alpha=1.0, line_dash="dashed")
        
        # Man-Whitney U for Table (round and use scienitic notation if p-value > 0.001)
        manw_pval = scipy.stats.mannwhitneyu(Yscore_train[Ytrue_train==0], Yscore_train[Ytrue_train==1])[1]
        if manw_pval > 0.001:
            manw_pval_round = "%0.2f" % manw_pval
        else:
            manw_pval_round = "%0.2e" % manw_pval
        if testset is not None:
            testmanw_pval = scipy.stats.mannwhitneyu(Yscore_test[Ytrue_test==0], Yscore_test[Ytrue_test==1])[1]
            if testmanw_pval > 0.001:
                testmanw_pval_round = "%0.2f" % testmanw_pval
            else:
                testmanw_pval_round = "%0.2e" % testmanw_pval
        
        # Create a stats table for test 
        if testset is not None:
            teststats = binary_metrics(Ytrue_test, Yscore_test, cut_off=stats['val_cutoffscore'])
            teststats_round = {}
            for i in teststats.keys():
                teststats_round[i] = np.round(teststats[i],2)
            
        # Round stats, and stats_bootci for Table
        stats_round = {} 
        for i in stats.keys():
            stats_round[i] = np.round(stats[i],2)
        bootci_round = {} 
        for i in stats_bootci.keys():
            bootci_round[i] = np.round(stats_bootci[i],2)
        
        # Create table 
        tabledata = dict(evaluate=[["Train"]],
                         manw_pval=[["{}".format(manw_pval_round)]],
                         auc=[["{} ({}, {})".format(stats_round['auc'], bootci_round['auc'][0], bootci_round['auc'][1])]],
                         accuracy=[["{} ({}, {})".format(stats_round['accuracy'], bootci_round['accuracy'][0], bootci_round['accuracy'][1])]],
                         precision=[["{} ({}, {})".format(stats_round['precision'], bootci_round['precision'][0], bootci_round['precision'][1])]],
                         sensitivity=[["{} ({}, {})".format(stats_round['sensitivity'], bootci_round['sensitivity'][0], bootci_round['sensitivity'][1])]],
                         specificity=[["{} ({}, {})".format(stats_round['specificity'], bootci_round['specificity'][0], bootci_round['specificity'][1])]],
                         F1score=[["{} ({}, {})".format(stats_round['F1score'], bootci_round['F1score'][0], bootci_round['F1score'][1])]],
                         R2=[["{} ({}, {})".format(stats_round['R2'], bootci_round['R2'][0], bootci_round['R2'][1])]])
        
        # Append test data 
        if testset is not None:
            tabledata['evaluate'].append(["Test"])
            tabledata['manw_pval'].append([testmanw_pval_round]) 
            tabledata['auc'].append([teststats_round['auc']])
            tabledata['accuracy'].append([teststats_round['accuracy']])
            tabledata['precision'].append([teststats_round['precision']])
            tabledata['sensitivity'].append([teststats_round['sensitivity']])
            tabledata['specificity'].append([teststats_round['specificity']])
            tabledata['F1score'].append([teststats_round['F1score']])
            tabledata['R2'].append([teststats_round['R2']])
        
        # Plot table
        source = ColumnDataSource(data=tabledata)
        columns = [
            TableColumn(field="evaluate", title="Evaluate"),
            TableColumn(field="manw_pval", title="MW-U Pvalue"),
            TableColumn(field="R2", title="R2"),
            TableColumn(field="auc", title="AUC"),
            TableColumn(field="accuracy", title="Accuracy"),
            TableColumn(field="precision", title="Precision"),
            TableColumn(field="sensitivity", title="Sensitivity"),
            TableColumn(field="F1score", title="F1score")]
        table_bokeh = widgetbox(DataTable(source=source, columns=columns, width=950, height=90), width=950, height=80)
        
        # Title 
        if specificity is not False:
            title = "Specificity set to: {}".format(np.round(val, 2))
        elif cutoffscore is not False:
            title = "Score cut-off set to: {}".format(np.round(val, 2))
        else:
            title = "Specificity set to: {}".format(np.round(val, 2))
        title_bokeh = "<h3>{}</h3>".format(title)

        # Combine table, violin plot and roc plot into one figure
        fig = gridplot([[violin_bokeh, dist_bokeh, roc_bokeh], [table_bokeh]], toolbar_location="right")
        output_notebook()
        show(column(Div(text=title_bokeh, width=900, height=50), fig))

    def calc_bootci(self, bootnum=1000, type="perc"):
        """Calculates bootstrap confidence intervals based on bootlist."""
        bootlist = self.bootlist
        if type is "bca":
            boot = BCA(self, self.X, self.Y, self.bootlist, bootnum=bootnum)
        if type is "bc":
            boot = BC(self, self.X, self.Y, self.bootlist, bootnum=bootnum)
        if type is "perc":
            boot = Perc(self, self.X, self.Y, self.bootlist, bootnum=bootnum)
        self.bootci = boot.run()

    def plot_featureimportance(self, PeakTable, peaklist=None, ylabel="Label", sort=True):
        """Plots feature importance metrics."""
        if not hasattr(self, "bootci"):
            print("Use method calc_bootci prior to plot_featureimportance to add 95% confidence intervals to plots.")
            ci_coef = None
            ci_vip = None
        else:
            ci_coef = self.bootci["model.coef_"]
            ci_vip = self.bootci["model.vip_"]

        # Remove rows from PeakTable if not in peaklist
        if peaklist is not None:
            PeakTable = PeakTable[PeakTable["Name"].isin(peaklist)]
        peaklabel = PeakTable[ylabel]

        # Plot
        fig_1 = scatterCI(self.model.coef_, ci=ci_coef, label=peaklabel, hoverlabel=PeakTable[["Idx", "Name", "Label"]], hline=0, col_hline=True, title="Coefficient Plot", sort_abs=sort)
        fig_2 = scatterCI(self.model.vip_, ci=ci_vip, label=peaklabel, hoverlabel=PeakTable[["Idx", "Name", "Label"]], hline=1, col_hline=False, title="Variable Importance in Projection (VIP)", sort_abs=sort)
        fig = layout([[fig_1], [fig_2]])
        output_notebook()
        show(fig)

        # Return table with: Idx, Name, Label, Coefficient, 95CI, VIP, 95CI
        coef = pd.DataFrame([self.model.coef_, self.bootci["model.coef_"]]).T
        coef.rename(columns={0: "Coef", 1: "Coef-95CI"}, inplace=True)
        vip = pd.DataFrame([self.model.vip_, self.bootci["model.vip_"]]).T
        vip.rename(columns={0: "VIP", 1: "VIP-95CI"}, inplace=True)

        Peaksheet = PeakTable.copy()
        Peaksheet["Coef"] = coef["Coef"].values
        Peaksheet["Coef-95CI"] = coef["Coef-95CI"].values
        Peaksheet["VIP"] = vip["VIP"].values
        Peaksheet["VIP-95CI"] = vip["VIP-95CI"].values
        return Peaksheet

    def permutation_test(self, nperm=100):
        """Plots permutation test figures."""
        fig = permutation_test(self, self.X, self.Y, nperm=nperm)
        output_notebook()
        show(fig)
        
    @staticmethod
    def _get_sens_spec(Ytrue, Yscore, cuttoff_val):
        Yscore_round = np.where(np.array(Yscore) > cuttoff_val, 1, 0) 
        tn, fp, fn, tp = metrics.confusion_matrix(Ytrue, Yscore_round).ravel()
        sensitivity  = tp / (tp+fn)
        specificity  = tn / (tn+fp)
        return sensitivity, specificity
    
    @staticmethod
    def _get_sens_cuttoff(Ytrue, Yscore, specificity_val):
        fpr0 = 1- specificity_val
        fpr, sensitivity , thresholds = metrics.roc_curve(Ytrue, Yscore, pos_label=1)
        idx = np.abs(fpr - fpr0).argmin() # this find the closest value in fpr to fpr0
        return sensitivity[idx], thresholds[idx]
    
    @staticmethod
    def _get_sens_spec_cuttoff(Ytrue, Yscore):
        fpr, sensitivity , thresholds = metrics.roc_curve(Ytrue, Yscore, pos_label=1)
        idx = (1 - fpr + sensitivity).argmax() # Using youden index, find max value
        specificity_idx = 1 - fpr[idx]
        return specificity_idx, sensitivity[idx], thresholds[idx]